/*
 * File:        test random gen
 * Author:      Bruce Land
 * Adapted from:
 *              main.c by
 * Author:      Syed Tahmid Mahbub
 * Target PIC:  PIC32MX250F128B
 */

// graphics libraries
#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
// need for rand function
#include <stdlib.h>

// threading library
#include <plib.h>
// config.h sets 40 MHz
#define	SYS_FREQ 40000000
#include "pt_cornell_TFT.h"

/* Demo code for interfacing TFT (ILI9340 controller) to PIC32
 * The library has been modified from a similar Adafruit library
 */
// Adafruit data:
/***************************************************
  This is an example sketch for the Adafruit 2.2" SPI display.
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

// string buffer
char buffer[60];

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_timer, pt_seed_init, pt_key ;

// system 1 second interval tick
int sys_time_seconds ;

// keypad valid press
int key_pressed ;

// randomized seed values generated by seed_init thread
short seed_short;
int seed_int;

// play white noise from PWM pin 21
volatile char play_state = 0 ;

// === linear feedback shift regiester =====================
#define bit30 0b01000000000000000000000000000000
#define bit27 0b00001000000000000000000000000000
int lfsr, bit0, bit1 ;
#define LFSR_SHIFT \
do { \
    bit0 = (lfsr >> 28) & 1 ; \
    bit1 = (lfsr >> 31) & 1 ;\
    lfsr = lfsr << 1 ;\
    lfsr = lfsr + (bit0 ^ bit1) ;\
} while(0) ;

// == fixed point 2.14 format ===============================
// resolution 2^-14 =  6.1035e-5
// dynamic range is +1.9999/-2.0
typedef signed short fix16 ;
#define multfix16(a,b) ((fix16)((((long)(a))*((long)(b)))>>14)) //multiply two fixed 2.14
#define float2fix16(a) ((fix16)((a)*16384.0)) // 2^14
#define fix2float16(a) ((float)(a)/16384.0)

//== One-pole Butterworth lowpass filter, 16-bit ==========================
// coeff = {b2, -a2} noting that b1=b2 and a1=1
// history = { last_input, last_output}
short filter_number; // 0 to 8 picks coeff second index
fix16 coeff[2][9], coeff_current[2], history[2], output, input ;
fix16 IIR_butter_1_16(fix16 input, fix16 *coeff, fix16 *history )
{
    fix16 output;
    output = multfix16(input+history[0], coeff[0]) + multfix16(history[1], coeff[1]) ;
    history[0] = input ;
    history[1] = output ;
    return output ;
}

// === Timer Thread =================================================
// update a 1 second tick counter
static PT_THREAD (protothread_timer(struct pt *pt))
{
    PT_BEGIN(pt);
     tft_setCursor(0, 0);
     tft_setTextColor(ILI9340_WHITE);  tft_setTextSize(1);
     tft_writeString("Time in seconds since boot\n");
      while(1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(1000) ;
        sys_time_seconds++ ;
        
        // draw sys_time
        tft_fillRoundRect(0,10, 100, 14, 1, ILI9340_BLACK);// x,y,w,h,radius,color
        tft_setCursor(0, 10);
        tft_setTextColor(ILI9340_YELLOW); tft_setTextSize(1);
        sprintf(buffer,"%d", sys_time_seconds);
        tft_writeString(buffer);

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // timer thread


// === random seed thread =============================================
// use floating ADC channel as a seed
static PT_THREAD (protothread_seed_init(struct pt *pt))
{
    PT_BEGIN(pt);
        static int adc_seed ;
        adc_seed =  PT_GET_TIME() + 1 ; // just to make sure it is not zero
  
        // AN9 pin 26
        // UNTERMINATED, FLOATING INPUT
        AcquireADC10(); // start ADC sampling -- connects ADC sample cap to circuit
        // wait for complete
        while (!AD1CON1bits.DONE){}; // Wait for ADC conversion
        // read the result of channel from the idle buffer
        adc_seed += ReadADC10(0) ;

        // wait a millisecond to de-correlate ADC readings
        PT_YIELD_TIME_msec(1);
        // shift 5 to make 15 bits
        AcquireADC10(); // start ADC sampling -- connects ADC sample cap to circuit
        // wait for complete
        while (!AD1CON1bits.DONE){}; // Wait for ADC conversion
        // read the result of channel from the idle buffer
        // overlap low order and high order bits
         adc_seed += (ReadADC10(0)<<5) ;

        // used the ADC sum as an input-seed and use rand to give final seed
        srand(adc_seed);
        seed_short = (short)(rand() & 0x7fff);
        seed_int =  rand() & 0x7fffffff ;
        
        // Return to spawning thread, in this case the keypad thread
        PT_EXIT(pt);
  PT_END(pt);
} // seed generator thread

// === Keypad Thread =============================================
// connections:
// A0 -- row 1 -- thru 300 ohm resistor -- avoid short when two buttons pushed
// A1 -- row 2 -- thru 300 ohm resistor
// A2 -- row 3 -- thru 300 ohm resistor
// A3 -- row 4 -- thru 300 ohm resistor
// B7 -- col 1 -- 10k pulldown resistor -- avoid open circuit input when no button pushed
// B8 -- col 2 -- 10k pulldown resistor
// B9 -- col 3 -- 10k pulldown resistor

static PT_THREAD (protothread_key(struct pt *pt))
{
    PT_BEGIN(pt);
    static int keypad, i, j, pattern;
    static short histogram[240], x;
    static int cycles;
    // order is 0 thru 9 then * ==10 and # ==11
    // no press = -1
    // table is decoded to natural digit order (except for * and #)
    // 0x80 for col 1 ; 0x100 for col 2 ; 0x200 for col 3
    // 0x01 for row 1 ; 0x02 for row 2; etc
    static int keytable[12]={0x108, 0x81, 0x101, 0x201, 0x82, 0x102, 0x202, 0x84, 0x104, 0x204, 0x88, 0x208};
    // init the keypad pins A0-A3 and B7-B9
    // PortA ports as digital outputs
    mPORTASetPinsDigitalOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);    //Set port as output
    // PortB as inputs
    mPORTBSetPinsDigitalIn(BIT_7 | BIT_8 | BIT_9);    //Set port as input
    // turn on port B pulldown resistors bits 7, 8, 9
    //CNPDB = 0b1110000000; //0x00380 ; // for the three pins used for the keypad
    //CNPDB = CNPDB7 | CNPDB8 | CNPDB9 ; 
    //CNPUB = 0b1000000000; //0x00380 ; // for the three pins used for the keypad
    
      while(1) {

        // read each row sequentially
        mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
        pattern = 1; mPORTASetBits(pattern);
        
        // yield time
        PT_YIELD_TIME_msec(30);
        //mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
        //pattern = 1; mPORTASetBits(pattern);
        for (i=0; i<4; i++) {
            keypad  = mPORTBReadBits(BIT_7 | BIT_8 | BIT_9);
            if(keypad!=0) {keypad |= pattern ; break;}
            mPORTAClearBits(pattern);
            pattern <<= 1;
            mPORTASetBits(pattern);
        }

        // search for keycode
        if (keypad > 0){ // then button is pushed
            for (i=0; i<12; i++){
                if (keytable[i]==keypad) break;
            }
        }
        else i = -1; // no button pushed
        key_pressed = i;

        // interpret the button push
        // 1=start noise; 0=stop noise; #=new seed
        // new seed spawns a task
        if (key_pressed == 1) { // digit 1

            // profile the compute time
            //cycles = ReadTimer3();
            LFSR_SHIFT
            // profile the compute time
            //cycles = ReadTimer3() - cycles ;

            x = lfsr & 0x7fff ;
            for (j=0; j<240; j++){
                if (x >= j*140 && x<(j+1)*140) {
                    histogram[j]++ ;
                    tft_drawFastVLine(j, 80, histogram[j], ILI9340_WHITE);
                }
            } // for (j=0; j<240; j++){
            play_state = 1;
        } //if(key_pressed == 1)

        // stop the noise!
        if (key_pressed == 0) { // digit 0
            play_state = 0;
        } //i

        // new seed and clear histogram
        if (key_pressed == 11) { // '#'
            PT_SPAWN(pt, &pt_seed_init, protothread_seed_init(&pt_seed_init) );
            // use new seed to start pseudorandom number sequence
            //srand(seed_short);
            lfsr = seed_int;
            srand(seed_int)  ;
            for (j=0; j<240; j++){
                histogram[j] = 0 ;
            } ////for (j=0; j<240; j++){
             tft_fillScreen(ILI9340_BLACK);
         } //if (key_pressed == 11)

        // set the digitial filter cutoff frequency
        if (key_pressed>0 && key_pressed<10 ){
            coeff_current[0] =  coeff[0][key_pressed - 1] ;
            coeff_current[1] =  coeff[1][key_pressed - 1] ;
        }

        //tft_fillRoundRect(20,200, 100, 14, 1, ILI9340_BLACK);// x,y,w,h,radius,color
        //tft_setCursor(20, 200);
        //tft_setTextColor(ILI9340_WHITE); tft_setTextSize(1);
        //sprintf(buffer,"%d", cycles-1);
        //tft_writeString(buffer);

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // keypad thread

// == Timer 2 ISR =====================================================
// noise generator based on seed value
volatile short PWM_value ;
int filter_state ;
// PWM sample rate is 40 MHz/1024 or 39062 Hz
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2ISR(void)
{
    // implement the linear feedback shift register
    // xor bits 31 and 28 and put in low order bit
    // bandwidth is about (sample rate)/7
    //bit0 = (lfsr >> 28) & 1 ;
    //bit1 = (lfsr >> 31) & 1 ;
    //lfsr = lfsr << 1 ;
    //lfsr = lfsr + (bit0 ^ bit1) ;
    LFSR_SHIFT

    // add another bit to decorrelate output for higher bandwidth
    // bandwidth is about (sample rate)/3
    if (key_pressed == 2 | key_pressed == 3  ){
        LFSR_SHIFT
    }

    // add two more bits to decorrelate output for higher bandwidth
    // bandwidth is about (sample rate)/2
    if ( key_pressed == 3  ){
        LFSR_SHIFT
        LFSR_SHIFT
    }

    // lowpass filter the noise 2.14 bit
    // truncate RAND  value to 2.14 fixed fraction format
    output = IIR_butter_1_16((fix16)lfsr & 0x3fff, coeff_current, history );
    // PWM output: shift 4 to convert 14 bit fraction to 10 bit
    //PWM_value = (output>>4) & 0x3ff ;

    // raw lfsr output, unflitered
    PWM_value = lfsr & 0x3ff ;

    if (play_state) SetDCOC3PWM(PWM_value);
    else SetDCOC3PWM(0x2ff);
    // clear the timer interrupt flag
    mT2ClearIntFlag();
}

// === Main  ======================================================
void main(void) {
 SYSTEMConfigPerformance(PBCLK);
  
  ANSELA = 0; ANSELB = 0; CM1CON = 0; CM2CON = 0;

  // === Config timer and output compares to make pulses ========
  // set up timer2 to generate the wave period 10-bit resolution
  OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 1024);
  ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
  mT2ClearIntFlag(); // and clear the interrupt flag

  // set up compare3 for PWM mode
  OpenOC3(OC_ON | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE , 0x1ff, 0x1ff); //1/2 full scale
  // OC3 is PPS group 4, map to RPB10 (pin 21)
  PPSOutput(4, RPB10, OC3);

  // === Config timer for profiling code ========
  // set up timer3
  OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_1, 0xffff);

  // === set up filter ===========================
  // set up the IIR filter as output
  // from butter(1,0.005) in matlab (or Octave)
  coeff[0][0] = float2fix16(0.0078) ;
  coeff[1][0] = float2fix16(0.9844) ;
  // from butter(1,0.01) in matlab (or Octave)
  coeff[0][1] = float2fix16(0.0155) ;
  coeff[1][1] = float2fix16(0.9691) ;
  // from butter(1,0.02) in matlab (or Octave)
  coeff[0][2] = float2fix16(0.0305) ;
  coeff[1][2] = float2fix16(0.9391) ;
  // from butter(1,0.05) in matlab (or Octave)
  coeff[0][3] = float2fix16(0.0730) ;
  coeff[1][3] = float2fix16(0.8541) ;
  // from butter(1,0.1) in matlab (or Octave)
  coeff[0][4] = float2fix16(0.1367) ;
  coeff[1][4] = float2fix16(0.7265) ;
  // from butter(1,0.2) in matlab (or Octave)
  coeff[0][5] = float2fix16(0.2452) ;
  coeff[1][5] = float2fix16(0.5095) ;
  // from butter(1,0.5) in matlab (or Octave)
  coeff[0][6] = float2fix16(0.5) ;
  coeff[1][6] = float2fix16(0.0) ;

  history[0] = float2fix16(0.0) ;
  history[1] = float2fix16(0.0) ;

  // set up the ADC ///////////////////////////////////////
        // configure and enable the ADC
	CloseADC10();	// ensure the ADC is off before setting the configuration

	// define setup parameters for OpenADC10
	// Turn module on | ouput in integer | trigger mode auto | enable autosample
        // ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
        //  ADC_CLK_MANUAL -- turn off auto convert
        // ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
        // ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
        #define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_OFF //

	// define setup parameters for OpenADC10
	// ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
	#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
        //
	// Define setup parameters for OpenADC10
        // use peripherial bus clock | set sample time | set ADC clock divider
        // ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
        // ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
        #define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_15 | ADC_CONV_CLK_Tcy //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy

	// define setup parameters for OpenADC10
	// set AN11 and  as analog inputs
	#define PARAM4	ENABLE_AN9_ANA

	// define setup parameters for OpenADC10
	// do not assign channels to scan
	#define PARAM5	SKIP_SCAN_ALL

	// use ground as neg ref for A | use AN11 for input A
	// configure to sample AN11
	SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN9 ); // configure to sample AN11
	OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters defined above

	EnableADC10(); // Enable the ADC

  // === config threads ==========
  // turns OFF UART support and debugger pin
  PT_setup();

  // === setup system wide interrupts  ========
  INTEnableSystemMultiVectoredInt();

  // init the threads
  PT_INIT(&pt_timer);
  PT_INIT(&pt_seed_init);
  PT_INIT(&pt_key);

  // init the display
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
  //240x320 vertical display
  tft_setRotation(0); // Use tft_setRotation(1) for 320x240

  // seed random color
  srand(1);

  // round-robin scheduler for threads
  while (1){
      PT_SCHEDULE(protothread_timer(&pt_timer));
      PT_SCHEDULE(protothread_key(&pt_key));
      }
  } // main

// === end  ======================================================

